@@ -12,9 +12,8 @@
 // OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 // CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 
-use crate::{arithmetic::montgomery::*, c, error, limb::*};
+use crate::{arithmetic::limbs_from_hex, arithmetic::montgomery::*, c, error, limb::*};
 use core::marker::PhantomData;
-use untrusted;
 
 pub use self::elem::*;
 
@@ -39,23 +38,19 @@ pub struct Point {
     // `ops.num_limbs` elements are the X coordinate, the next
     // `ops.num_limbs` elements are the Y coordinate, and the next
     // `ops.num_limbs` elements are the Z coordinate. This layout is dictated
-    // by the requirements of the GFp_nistz256 code.
+    // by the requirements of the nistz256 code.
     xyz: [Limb; 3 * MAX_LIMBS],
 }
 
 impl Point {
-    pub fn new_at_infinity() -> Point {
-        Point {
+    pub fn new_at_infinity() -> Self {
+        Self {
             xyz: [0; 3 * MAX_LIMBS],
         }
     }
 }
 
-static ONE: Elem<Unencoded> = Elem {
-    limbs: limbs![1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
-    m: PhantomData,
-    encoding: PhantomData,
-};
+const ONE: Elem<Unencoded> = Elem::from_hex("1");
 
 /// Operations and values needed by all curve operations.
 pub struct CommonOps {
@@ -67,7 +62,6 @@ pub struct CommonOps {
     pub b: Elem<R>,
 
     // In all cases, `r`, `a`, and `b` may all alias each other.
-    elem_add_impl: unsafe extern "C" fn(r: *mut Limb, a: *const Limb, b: *const Limb),
     elem_mul_mont: unsafe extern "C" fn(r: *mut Limb, a: *const Limb, b: *const Limb),
     elem_sqr_mont: unsafe extern "C" fn(r: *mut Limb, a: *const Limb),
 
@@ -77,7 +71,12 @@ pub struct CommonOps {
 impl CommonOps {
     #[inline]
     pub fn elem_add<E: Encoding>(&self, a: &mut Elem<E>, b: &Elem<E>) {
-        binary_op_assign(self.elem_add_impl, a, b)
+        let num_limbs = self.num_limbs;
+        limbs_add_assign_mod(
+            &mut a.limbs[..num_limbs],
+            &b.limbs[..num_limbs],
+            &self.q.p[..num_limbs],
+        );
     }
 
     #[inline]
@@ -271,10 +270,7 @@ pub struct PublicScalarOps {
     pub scalar_ops: &'static ScalarOps,
     pub public_key_ops: &'static PublicKeyOps,
 
-    // XXX: `PublicScalarOps` shouldn't depend on `PrivateKeyOps`, but it does
-    // temporarily until `twin_mul` is rewritten.
-    pub private_key_ops: &'static PrivateKeyOps,
-
+    pub twin_mul: fn(g_scalar: &Scalar, p_scalar: &Scalar, p_xy: &(Elem<R>, Elem<R>)) -> Point,
     pub q_minus_n: Elem<Unencoded>,
 }
 
@@ -288,24 +284,15 @@ impl PublicScalarOps {
         }
     }
 
-    pub fn elem_equals(&self, a: &Elem<Unencoded>, b: &Elem<Unencoded>) -> bool {
-        for i in 0..self.public_key_ops.common.num_limbs {
-            if a.limbs[i] != b.limbs[i] {
-                return false;
-            }
-        }
-        true
+    pub fn elem_equals_vartime(&self, a: &Elem<Unencoded>, b: &Elem<Unencoded>) -> bool {
+        a.limbs[..self.public_key_ops.common.num_limbs]
+            == b.limbs[..self.public_key_ops.common.num_limbs]
     }
 
     pub fn elem_less_than(&self, a: &Elem<Unencoded>, b: &Elem<Unencoded>) -> bool {
         let num_limbs = self.public_key_ops.common.num_limbs;
         limbs_less_than_limbs_vartime(&a.limbs[..num_limbs], &b.limbs[..num_limbs])
     }
-
-    #[inline]
-    pub fn elem_sum(&self, a: &Elem<Unencoded>, b: &Elem<Unencoded>) -> Elem<Unencoded> {
-        binary_op(self.public_key_ops.common.elem_add_impl, a, b)
-    }
 }
 
 #[allow(non_snake_case)]
@@ -315,6 +302,19 @@ pub struct PrivateScalarOps {
     pub oneRR_mod_n: Scalar<RR>, // 1 * R**2 (mod n). TOOD: Use One<RR>.
 }
 
+// XXX: Inefficient and unnecessarily depends on `PrivateKeyOps`. TODO: implement interleaved wNAF
+// multiplication.
+fn twin_mul_inefficient(
+    ops: &PrivateKeyOps,
+    g_scalar: &Scalar,
+    p_scalar: &Scalar,
+    p_xy: &(Elem<R>, Elem<R>),
+) -> Point {
+    let scaled_g = ops.point_mul_base(g_scalar);
+    let scaled_p = ops.point_mul(p_scalar, p_xy);
+    ops.common.point_sum(&scaled_g, &scaled_p)
+}
+
 // This assumes n < q < 2*n.
 pub fn elem_reduced_to_scalar(ops: &CommonOps, elem: &Elem<Unencoded>) -> Scalar<Unencoded> {
     let num_limbs = ops.num_limbs;
@@ -426,7 +426,7 @@ fn parse_big_endian_fixed_consttime<M>(
     Ok(r)
 }
 
-extern "C" {
+prefixed_extern! {
     fn LIMBS_add_mod(
         r: *mut Limb,
         a: *const Limb,
@@ -438,10 +438,10 @@ extern "C" {
 
 #[cfg(test)]
 mod tests {
+    extern crate alloc;
     use super::*;
     use crate::test;
     use alloc::{format, vec, vec::Vec};
-    use untrusted;
 
     const ZERO_SCALAR: Scalar = Scalar {
         limbs: [0; MAX_LIMBS],
@@ -491,11 +491,11 @@ mod tests {
             let b = consume_elem(cops, test_case, "b");
             let expected_sum = consume_elem(cops, test_case, "r");
 
-            let mut actual_sum = a.clone();
+            let mut actual_sum = a;
             ops.public_key_ops.common.elem_add(&mut actual_sum, &b);
             assert_limbs_are_equal(cops, &actual_sum.limbs, &expected_sum.limbs);
 
-            let mut actual_sum = b.clone();
+            let mut actual_sum = b;
             ops.public_key_ops.common.elem_add(&mut actual_sum, &a);
             assert_limbs_are_equal(cops, &actual_sum.limbs, &expected_sum.limbs);
 
@@ -503,17 +503,17 @@ mod tests {
         })
     }
 
-    // XXX: There's no `GFp_nistz256_sub` in *ring*; it's logic is inlined into
+    // XXX: There's no `p256_sub` in *ring*; it's logic is inlined into
     // the point arithmetic functions. Thus, we can't test it.
 
     #[test]
     fn p384_elem_sub_test() {
-        extern "C" {
-            fn GFp_p384_elem_sub(r: *mut Limb, a: *const Limb, b: *const Limb);
+        prefixed_extern! {
+            fn p384_elem_sub(r: *mut Limb, a: *const Limb, b: *const Limb);
         }
         elem_sub_test(
             &p384::COMMON_OPS,
-            GFp_p384_elem_sub,
+            p384_elem_sub,
             test_file!("ops/p384_elem_sum_tests.txt"),
         );
     }
@@ -554,17 +554,17 @@ mod tests {
         })
     }
 
-    // XXX: There's no `GFp_nistz256_div_by_2` in *ring*; it's logic is inlined
+    // XXX: There's no `p256_div_by_2` in *ring*; it's logic is inlined
     // into the point arithmetic functions. Thus, we can't test it.
 
     #[test]
     fn p384_elem_div_by_2_test() {
-        extern "C" {
-            fn GFp_p384_elem_div_by_2(r: *mut Limb, a: *const Limb);
+        prefixed_extern! {
+            fn p384_elem_div_by_2(r: *mut Limb, a: *const Limb);
         }
         elem_div_by_2_test(
             &p384::COMMON_OPS,
-            GFp_p384_elem_div_by_2,
+            p384_elem_div_by_2,
             test_file!("ops/p384_elem_div_by_2_tests.txt"),
         );
     }
@@ -590,27 +590,28 @@ mod tests {
         })
     }
 
-    // TODO: Add test vectors that test the range of values above `q`.
+    // There is no `ecp_nistz256_neg` on other targets.
+    #[cfg(target_arch = "x86_64")]
     #[test]
     fn p256_elem_neg_test() {
-        extern "C" {
-            fn GFp_nistz256_neg(r: *mut Limb, a: *const Limb);
+        prefixed_extern! {
+            fn ecp_nistz256_neg(r: *mut Limb, a: *const Limb);
         }
         elem_neg_test(
             &p256::COMMON_OPS,
-            GFp_nistz256_neg,
+            ecp_nistz256_neg,
             test_file!("ops/p256_elem_neg_tests.txt"),
         );
     }
 
     #[test]
     fn p384_elem_neg_test() {
-        extern "C" {
-            fn GFp_p384_elem_neg(r: *mut Limb, a: *const Limb);
+        prefixed_extern! {
+            fn p384_elem_neg(r: *mut Limb, a: *const Limb);
         }
         elem_neg_test(
             &p384::COMMON_OPS,
-            GFp_p384_elem_neg,
+            p384_elem_neg,
             test_file!("ops/p384_elem_neg_tests.txt"),
         );
     }
@@ -692,10 +693,10 @@ mod tests {
         test::run(test_file, |section, test_case| {
             assert_eq!(section, "");
             let cops = ops.common;
-            let mut a = consume_scalar(cops, test_case, "a");
+            let a = consume_scalar(cops, test_case, "a");
             let b = consume_scalar_mont(cops, test_case, "b");
             let expected_result = consume_scalar(cops, test_case, "r");
-            let actual_result = ops.scalar_product(&mut a, &b);
+            let actual_result = ops.scalar_product(&a, &b);
             assert_limbs_are_equal(cops, &actual_result.limbs, &expected_result.limbs);
 
             Ok(())
@@ -704,18 +705,18 @@ mod tests {
 
     #[test]
     fn p256_scalar_square_test() {
-        extern "C" {
-            fn GFp_p256_scalar_sqr_rep_mont(r: *mut Limb, a: *const Limb, rep: Limb);
+        prefixed_extern! {
+            fn p256_scalar_sqr_rep_mont(r: *mut Limb, a: *const Limb, rep: Limb);
         }
         scalar_square_test(
             &p256::SCALAR_OPS,
-            GFp_p256_scalar_sqr_rep_mont,
+            p256_scalar_sqr_rep_mont,
             test_file!("ops/p256_scalar_square_tests.txt"),
         );
     }
 
     // XXX: There's no `p384_scalar_square_test()` because there's no dedicated
-    // `GFp_p384_scalar_sqr_rep_mont()`.
+    // `p384_scalar_sqr_rep_mont()`.
 
     fn scalar_square_test(
         ops: &ScalarOps,
@@ -794,8 +795,8 @@ mod tests {
 
     #[test]
     fn p256_point_sum_mixed_test() {
-        extern "C" {
-            fn GFp_nistz256_point_add_affine(
+        prefixed_extern! {
+            fn p256_point_add_affine(
                 r: *mut Limb,   // [p256::COMMON_OPS.num_limbs*3]
                 a: *const Limb, // [p256::COMMON_OPS.num_limbs*3]
                 b: *const Limb, // [p256::COMMON_OPS.num_limbs*2]
@@ -803,12 +804,12 @@ mod tests {
         }
         point_sum_mixed_test(
             &p256::PRIVATE_KEY_OPS,
-            GFp_nistz256_point_add_affine,
+            p256_point_add_affine,
             test_file!("ops/p256_point_sum_mixed_tests.txt"),
         );
     }
 
-    // XXX: There is no `GFp_nistz384_point_add_affine()`.
+    // XXX: There is no `nistz384_point_add_affine()`.
 
     fn point_sum_mixed_test(
         ops: &PrivateKeyOps,
@@ -839,30 +840,30 @@ mod tests {
 
     #[test]
     fn p256_point_double_test() {
-        extern "C" {
-            fn GFp_nistz256_point_double(
+        prefixed_extern! {
+            fn p256_point_double(
                 r: *mut Limb,   // [p256::COMMON_OPS.num_limbs*3]
                 a: *const Limb, // [p256::COMMON_OPS.num_limbs*3]
             );
         }
         point_double_test(
             &p256::PRIVATE_KEY_OPS,
-            GFp_nistz256_point_double,
+            p256_point_double,
             test_file!("ops/p256_point_double_tests.txt"),
         );
     }
 
     #[test]
     fn p384_point_double_test() {
-        extern "C" {
-            fn GFp_nistz384_point_double(
+        prefixed_extern! {
+            fn p384_point_double(
                 r: *mut Limb,   // [p384::COMMON_OPS.num_limbs*3]
                 a: *const Limb, // [p384::COMMON_OPS.num_limbs*3]
             );
         }
         point_double_test(
             &p384::PRIVATE_KEY_OPS,
-            GFp_nistz384_point_double,
+            p384_point_double,
             test_file!("ops/p384_point_double_tests.txt"),
         );
     }
@@ -978,6 +979,7 @@ mod tests {
     fn p256_point_mul_base_test() {
         point_mul_base_tests(
             &p256::PRIVATE_KEY_OPS,
+            |s| p256::PRIVATE_KEY_OPS.point_mul_base(s),
             test_file!("ops/p256_point_mul_base_tests.txt"),
         );
     }
@@ -986,16 +988,21 @@ mod tests {
     fn p384_point_mul_base_test() {
         point_mul_base_tests(
             &p384::PRIVATE_KEY_OPS,
+            |s| p384::PRIVATE_KEY_OPS.point_mul_base(s),
             test_file!("ops/p384_point_mul_base_tests.txt"),
         );
     }
 
-    fn point_mul_base_tests(ops: &PrivateKeyOps, test_file: test::File) {
+    pub(super) fn point_mul_base_tests(
+        ops: &PrivateKeyOps,
+        f: impl Fn(&Scalar) -> Point,
+        test_file: test::File,
+    ) {
         test::run(test_file, |section, test_case| {
             assert_eq!(section, "");
             let g_scalar = consume_scalar(ops.common, test_case, "g_scalar");
             let expected_result = consume_point(ops, test_case, "r");
-            let actual_result = ops.point_mul_base(&g_scalar);
+            let actual_result = f(&g_scalar);
             assert_point_actual_equals_expected(ops, &actual_result, &expected_result);
             Ok(())
         })
@@ -1007,25 +1014,25 @@ mod tests {
         expected_point: &TestPoint,
     ) {
         let cops = ops.common;
-        let actual_x = &cops.point_x(&actual_point);
-        let actual_y = &cops.point_y(&actual_point);
-        let actual_z = &cops.point_z(&actual_point);
+        let actual_x = &cops.point_x(actual_point);
+        let actual_y = &cops.point_y(actual_point);
+        let actual_z = &cops.point_z(actual_point);
         match expected_point {
             TestPoint::Infinity => {
                 let zero = Elem::zero();
-                assert_elems_are_equal(cops, &actual_z, &zero);
+                assert_elems_are_equal(cops, actual_z, &zero);
             }
             TestPoint::Affine(expected_x, expected_y) => {
-                let zz_inv = ops.elem_inverse_squared(&actual_z);
-                let x_aff = cops.elem_product(&actual_x, &zz_inv);
+                let zz_inv = ops.elem_inverse_squared(actual_z);
+                let x_aff = cops.elem_product(actual_x, &zz_inv);
                 let y_aff = {
                     let zzzz_inv = cops.elem_squared(&zz_inv);
-                    let zzz_inv = cops.elem_product(&actual_z, &zzzz_inv);
-                    cops.elem_product(&actual_y, &zzz_inv)
+                    let zzz_inv = cops.elem_product(actual_z, &zzzz_inv);
+                    cops.elem_product(actual_y, &zzz_inv)
                 };
 
-                assert_elems_are_equal(cops, &x_aff, &expected_x);
-                assert_elems_are_equal(cops, &y_aff, &expected_y);
+                assert_elems_are_equal(cops, &x_aff, expected_x);
+                assert_elems_are_equal(cops, &y_aff, expected_y);
             }
         }
     }
@@ -1065,7 +1072,7 @@ mod tests {
         p
     }
 
-    fn consume_point_elem(ops: &CommonOps, limbs_out: &mut [Limb], elems: &Vec<&str>, i: usize) {
+    fn consume_point_elem(ops: &CommonOps, limbs_out: &mut [Limb], elems: &[&str], i: usize) {
         let bytes = test::from_hex(elems[i]).unwrap();
         let bytes = untrusted::Input::from(&bytes);
         let r: Elem<Unencoded> = elem_parse_big_endian_fixed_consttime(ops, bytes).unwrap();
@@ -1080,7 +1087,7 @@ mod tests {
     }
 
     fn consume_point(ops: &PrivateKeyOps, test_case: &mut test::TestCase, name: &str) -> TestPoint {
-        fn consume_point_elem(ops: &CommonOps, elems: &Vec<&str>, i: usize) -> Elem<R> {
+        fn consume_point_elem(ops: &CommonOps, elems: &[&str], i: usize) -> Elem<R> {
             let bytes = test::from_hex(elems[i]).unwrap();
             let bytes = untrusted::Input::from(&bytes);
             let unencoded: Elem<Unencoded> =
@@ -1113,15 +1120,19 @@ mod tests {
         actual: &[Limb; MAX_LIMBS],
         expected: &[Limb; MAX_LIMBS],
     ) {
-        for i in 0..ops.num_limbs {
-            if actual[i] != expected[i] {
-                let mut s = alloc::string::String::new();
-                for j in 0..ops.num_limbs {
-                    let formatted = format!("{:016x}", actual[ops.num_limbs - j - 1]);
-                    s.push_str(&formatted);
-                }
-                panic!("Actual != Expected,\nActual = {}", s);
+        if actual[..ops.num_limbs] != expected[..ops.num_limbs] {
+            let mut actual_s = alloc::string::String::new();
+            let mut expected_s = alloc::string::String::new();
+            for j in 0..ops.num_limbs {
+                let formatted = format!("{:016x}", actual[ops.num_limbs - j - 1]);
+                actual_s.push_str(&formatted);
+                let formatted = format!("{:016x}", expected[ops.num_limbs - j - 1]);
+                expected_s.push_str(&formatted);
             }
+            panic!(
+                "Actual != Expected,\nActual = {}, Expected = {}",
+                actual_s, expected_s
+            );
         }
     }
 
@@ -1171,84 +1182,6 @@ mod tests {
     }
 }
 
-#[cfg(feature = "internal_benches")]
-mod internal_benches {
-    use super::{Limb, MAX_LIMBS};
-
-    pub const LIMBS_1: [Limb; MAX_LIMBS] = limbs![1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0];
-
-    pub const LIMBS_ALTERNATING_10: [Limb; MAX_LIMBS] = limbs![
-        0b10101010_10101010_10101010_10101010,
-        0b10101010_10101010_10101010_10101010,
-        0b10101010_10101010_10101010_10101010,
-        0b10101010_10101010_10101010_10101010,
-        0b10101010_10101010_10101010_10101010,
-        0b10101010_10101010_10101010_10101010,
-        0b10101010_10101010_10101010_10101010,
-        0b10101010_10101010_10101010_10101010,
-        0b10101010_10101010_10101010_10101010,
-        0b10101010_10101010_10101010_10101010,
-        0b10101010_10101010_10101010_10101010,
-        0b10101010_10101010_10101010_10101010
-    ];
-}
-
-#[cfg(feature = "internal_benches")]
-macro_rules! bench_curve {
-    ( $vectors:expr ) => {
-        use super::super::{Elem, Scalar};
-        extern crate test;
-
-        #[bench]
-        fn elem_inverse_squared_bench(bench: &mut test::Bencher) {
-            // This benchmark assumes that `elem_inverse_squared()` is
-            // constant-time so inverting 1 mod q is as good of a choice as
-            // anything.
-            let mut a = Elem::zero();
-            a.limbs[0] = 1;
-            bench.iter(|| {
-                let _ = PRIVATE_KEY_OPS.elem_inverse_squared(&a);
-            });
-        }
-
-        #[bench]
-        fn elem_product_bench(bench: &mut test::Bencher) {
-            // This benchmark assumes that the multiplication is constant-time
-            // so 0 * 0 is as good of a choice as anything.
-            let a: Elem<R> = Elem::zero();
-            let b: Elem<R> = Elem::zero();
-            bench.iter(|| {
-                let _ = COMMON_OPS.elem_product(&a, &b);
-            });
-        }
-
-        #[bench]
-        fn elem_squared_bench(bench: &mut test::Bencher) {
-            // This benchmark assumes that the squaring is constant-time so
-            // 0**2 * 0 is as good of a choice as anything.
-            let a = Elem::zero();
-            bench.iter(|| {
-                let _ = COMMON_OPS.elem_squared(&a);
-            });
-        }
-
-        #[bench]
-        fn scalar_inv_to_mont_bench(bench: &mut test::Bencher) {
-            const VECTORS: &[Scalar] = $vectors;
-            let vectors_len = VECTORS.len();
-            let mut i = 0;
-            bench.iter(|| {
-                let _ = SCALAR_OPS.scalar_inv_to_mont(&VECTORS[i]);
-
-                i += 1;
-                if i == vectors_len {
-                    i = 0;
-                }
-            });
-        }
-    };
-}
-
 mod elem;
 pub mod p256;
 pub mod p384;
