@@ -15,26 +15,34 @@
 //! Polyfills for functionality that will (hopefully) be added to Rust's
 //! standard library soon.
 
-#[macro_use]
-pub mod convert;
-
 #[inline(always)]
 pub const fn u64_from_usize(x: usize) -> u64 {
     x as u64
 }
 
-#[inline(always)]
 pub fn usize_from_u32(x: u32) -> usize {
     x as usize
 }
 
-pub mod slice {
-    // https://github.com/rust-lang/rust/issues/27750
-    // https://internals.rust-lang.org/t/stabilizing-basic-functions-on-arrays-and-slices/2868
-    #[inline(always)]
-    pub fn fill(dest: &mut [u8], value: u8) {
-        for d in dest {
-            *d = value;
-        }
-    }
-}
+#[macro_use]
+mod chunks_fixed;
+
+mod array_flat_map;
+mod array_flatten;
+mod array_split_map;
+
+#[cfg(feature = "alloc")]
+mod leading_zeros_skipped;
+
+#[cfg(test)]
+mod test;
+
+mod unwrap_const;
+
+pub use self::{
+    array_flat_map::ArrayFlatMap, array_flatten::ArrayFlatten, array_split_map::ArraySplitMap,
+    chunks_fixed::*, unwrap_const::unwrap_const,
+};
+
+#[cfg(feature = "alloc")]
+pub use leading_zeros_skipped::LeadingZerosStripped;
